import numpy as np
import torch
import torch.distributed as dist

def safe_all_reduce(x, reduce_op=dist.ReduceOp.SUM):
    if dist.is_initialized():
        dist.all_reduce(x, reduce_op)
    return x


def safe_all_mean(x):
    x = safe_all_reduce(x)
    if dist.is_initialized():
        x /= dist.get_world_size()
    return x


def safe_all_gather(x, dim=0):
    if dist.is_initialized():
        xs = [torch.empty_like(x) for _ in range(dist.get_world_size())]
        dist.all_gather(xs, x)
        x = torch.cat(xs, dim=dim)
    return x


def safe_barrier():
    if dist.is_initialized():
        dist.barrier()


def safe_broadcast(x, src):
    if dist.is_initialized():
        dist.broadcast(x, src)


def safe_to_tensor(x, device="cpu"):
    non_blocking = device != "cpu"

    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).to(device, non_blocking=non_blocking)
    elif isinstance(x, torch.Tensor):
        return x.to(device, non_blocking=non_blocking)
    elif isinstance(x, (list, tuple)):
        return torch.tensor(x, device=device)
    elif isinstance(x, dict):
        return {k: safe_to_tensor(v, device=device) for k, v in x.items()}
    return x


def get_world_size():
    if not dist.is_available():
        return 1
    if not dist.is_initialized():
        return 1
    return dist.get_world_size()


def reduce_dict(input_dict, average=True):
    """
    Args:
        input_dict (dict): all the values will be reduced
        average (bool): whether to do average or sum
    Reduce the values in the dictionary from all processes so that all processes
    have the averaged results. Returns a dict with the same fields as
    input_dict, after reduction.
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict
    with torch.no_grad():
        names = []
        values = []
        # sort the keys so that they are consistent across processes
        for k in sorted(input_dict.keys()):
            names.append(k)
            value = input_dict[k]
            if not isinstance(value, torch.Tensor):
                value = torch.tensor(value).cuda()
            values.append(value)
        values = torch.stack(values, dim=0)
        dist.all_reduce(values)
        if average:
            values /= world_size
        reduced_dict = {k: v.item() if v.numel() == 1 else v for k, v in zip(names, values)}
    return reduced_dict